/* $Workfile: TecSs7.c $       $Revision: 6 $      $Modtime: 7/03/07 11:50a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X4SS7.C

    A simple example console application demonstrating the use of OTX Driver.

    The applications works with the following board combinations:

      Thor-2-ExpressCard
      Thor-4-ExpressCard

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures all available E1T1 devices on the board. The E1T1 spans should be
    looped with an external loop cable (Li0<->Li1).
    
    Li0 sends MSUs to Li1. Li1 acknowledges the MSU.
    Li1 sends MSUs to Li0. Li0 acknowledges the MSU.

    SS7 FISU filtering is enabled.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaDef.h"
#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_DEFAULT_SS7_SIM_FRAMES 300

#define DEMO_DEFAULT_LOG_FILE "TecSs7.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_HDLC_RX_BUFFER_SIZE 2048
#define DEMO_DEFAULT_HDLC_MAX_FRAME_LENGTH 272

#define DEMO_NUMBER_OF_DEVS 4

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL                bExitOnError;
  char                  *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  char                   cwd[OTX_MAX_PATH];
  int                    coDsp;
  int                    coLi;
  int                    nSimFrames;
  int                    cbMaxHdlcFrameLength;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bSs7FisuFiltering;
  OTX_BOOL               bHdlcBitInversion;
  int                    boardNo;
  OTX_BOOL               bUseDspDebugger;
  OTX_BOOL               b56kbpsChannel;
  int                    nTimeSlots;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxDspSpmE             eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;

void *DemoMsuTransmitFunction(
  void *lpParameter   // thread data
);

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
FILE       *g_pFrameRxFile;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs
OTX_BOOL    g_bHdlcVerbose = OTX_FALSE;
OTX_UINT32  g_coHdlcRxBytes[DEMO_NUMBER_OF_DEVS];
OTX_UINT32  g_coHdlcRxFrames[DEMO_NUMBER_OF_DEVS];
OTX_UINT32  g_coHdlcBadRxFrames[DEMO_NUMBER_OF_DEVS];

OTX_BOOL    g_bSs7SimulationActive[DEMO_NUMBER_OF_DEVS] = { 0 };
OTX_INT32   g_nSs7SimulationFramesReceived[DEMO_NUMBER_OF_DEVS] = { 0 };

OsaMutexS g_demoLogfCritSec;
OsaMutexS g_demoSendSs7DataCritSec;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hDualE1T1[1]; 
  OTX_HANDLE hLi[4];
  OTX_HANDLE hDsp[1];
} g_hPhysDevs;

struct {
  OTX_INT32  nDev;
  OTX_INT32  nRxDev;   // Device number that receives the frames that this sender is sending
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_INT32 coMsusSent;
  OTX_INT32 coFisusSent;

  OtxHdlcSS7FisuS fisu;

  OTX_INT32 busyCount;  // Should always be zero in this app

  OTX_INT32  nLi;
  OTX_INT32  nTs;
  OTX_INT32  nDsp;
} g_sqDspHdlcSender[DEMO_NUMBER_OF_DEVS];

struct {
  OTX_INT32  nDev;
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;

  OTX_UINT8  nPrevFsn;
  OTX_INT32  nSs7Errors;

  OTX_INT32  nLi;
  OTX_INT32  nTs;
  OTX_INT32  nDsp;
} g_sqDspHdlcReceiver[DEMO_NUMBER_OF_DEVS];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


#if defined(_WIN32_WCE)
#define CLOCKS_PER_SEC 1
typedef int clock_t;
int clock() { return 0; }
#endif 
/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func
    Initalize the global data for the HDLC devices

*******************************************************************************/
void DemoInitializeGlobalData(
  void
)
{
  OTX_INT32 i;
  OTX_INT32 nStartTs;

  assert((DEMO_NUMBER_OF_DEVS % 2) == 0);  // DEMO_NUMBER_OF_DEVS must be an even number since we spread the devices between the two Lis.
 
  if (DEMO_NUMBER_OF_DEVS == 2) {
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        nStartTs = 24 * 4  ;  //+ nLi * 2 See comment above regarding the timeslot mapping for the DSP highway on this board
      } else
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
        // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
        nStartTs = (24 * 4)  ;  //+ nLi See comment above regarding the timeslot mapping for the DSP highway on this board
      } else {
        nStartTs = 24;
      }        
    } else {
      // E1 mode
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        nStartTs = (16 * 4)  ;  // + (nLi * 2)See comment above regarding the timeslot mapping for the DSP highway on this board
      } else
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
        // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
        // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
        // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
        // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
        nStartTs = (16 * 4)  ;  // + nLi See comment above regarding the timeslot mapping for the DSP highway on this board
      } else {
        nStartTs = 16;
      }  
    }
  } else {
    nStartTs = 1; 
  }


  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    g_sqDspHdlcReceiver[i].nDev = i;
    g_sqDspHdlcReceiver[i].nLi  = i / (DEMO_NUMBER_OF_DEVS/2);
    g_sqDspHdlcReceiver[i].nTs  = nStartTs + (i % (DEMO_NUMBER_OF_DEVS/2));
    g_sqDspHdlcReceiver[i].nDsp = 0;                          

    g_sqDspHdlcSender[i].nDev   = i;
    g_sqDspHdlcSender[i].nLi    = i / (DEMO_NUMBER_OF_DEVS/2);;
    g_sqDspHdlcSender[i].nTs    = nStartTs + (i % (DEMO_NUMBER_OF_DEVS/2));; 
    g_sqDspHdlcSender[i].nDsp   = 0;                          
    g_sqDspHdlcSender[i].nRxDev = ((i + (DEMO_NUMBER_OF_DEVS/2)) % DEMO_NUMBER_OF_DEVS);             
    
    g_coHdlcRxFrames[i]         = 0;
    g_coHdlcRxBytes[i]          = 0;
    g_coHdlcBadRxFrames[i]      = 0;
    g_bSs7SimulationActive[i]   = OTX_FALSE;
  }

  // print out
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    DemoLogf("HDLC Device[%.2d]: Li=%.2d LiTs=%.2d Dsp=%.2d RxDev=%.2d\n", 
      i,
      g_sqDspHdlcSender[i].nLi,       
      g_sqDspHdlcSender[i].nTs,       
      g_sqDspHdlcSender[i].nDsp,      
      g_sqDspHdlcSender[i].nRxDev     
    );
  }

}




/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDualE1T1; 
  OTX_INT32  nQuadE1T1; 

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {
      // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a dual E1T1 device. 
      nDualE1T1 = 0;
      DemoLogf("Open OTX_DEVICE_DUAL_T1E1 #%d\n", nDualE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, nDualE1T1, nDualE1T1, g_hEventQueue, &g_hPhysDevs.hDualE1T1[nDualE1T1]); DemoCheckRc(result);
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nDualE1T1 = i/2;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_DUAL_T1E1 #%d)\n", i, nDualE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hDualE1T1[nDualE1T1], OTX_DEVICE_LI_T1E1, i%2, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This 
    // demo however needs the card to support the transmit data so the 
    // card will use 2 receivers and the 2 transmitters.
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
    DemoCheckRc(result);
    DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n");

    // we'll source the Falc from the DSP device, and source the DSP from the FALC
    //  FALC <-- DSP
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_FALC_SOURCED, 
                                           OTX_BRD_FALC_SOURCED_DSP); 
    DemoCheckRc(result);
    //  DSP <-- FALC
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_DSP_SOURCED, 
                                           OTX_BRD_DSP_SOURCED_FALC); 
    DemoCheckRc(result);

    DemoLogf("  Set FPGA highway path DSP->Falc and Falc->DSP\n");

    // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hDualE1T1[nQuadE1T1]); DemoCheckRc(result);
    }
    
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hDualE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Get handles to the physcial DSP devices
    i = 0;
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);
  }

	// Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXHDLC_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC_INTMEM\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2_INTMEM\n", i);
      break;
    case OTX_DSP_SPM_OTXSPM2EX_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2EX_INTMEM\n", i);
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(result);
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Initialize the T1/E1 Line Interface Devices

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);
  }

}

/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDualE1T1; 
  OTX_INT32  nQuadE1T1; 

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  // If board has Quad or Dual E1T1 devices then enable them
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {      
      // Enable Dual E1T1 device
      nDualE1T1 = 0;
      DemoLogf("Enable OTX_DEVICE_DUAL_T1E1 #%d device\n", nDualE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hDualE1T1[nDualE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    {
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hDualE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
    }
    break;
  default:
    // Nothing
    break;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_sqDspHdlcSender[i].hHandle); DemoCheckRc(result);
    
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_sqDspHdlcReceiver[i].hHandle); DemoCheckRc(result);

  }
  
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqLiTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_INT32 nTsLi;
  OTX_INT32 nTsDsp;
  OTX_INT32 nCapacity;
  OTX_UINT32 coTsPerDspHwy = g_nDspHighwaySpeed / 64;

  
  // 64kbps or 56kbps HDLC capacity
  if (g_demoEnv.b56kbpsChannel) {
    DemoLogf("Configuring one 56kbps HDLC channel per Li\n");
    nCapacity = 56;
  } else {
    DemoLogf("Configuring one 64kbps HDLC channel per Li\n");
    nCapacity = 64;
  }

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    nLi = g_sqDspHdlcReceiver[i].nLi;
    nDsp = g_sqDspHdlcReceiver[i].nDsp;
    
    nTsLi = g_sqDspHdlcReceiver[i].nTs; 
    
    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
      // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
      nTsDsp = (nTsLi * 4) + (nLi * 2) ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
    } else
    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
      // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
      // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
      // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
      // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
      nTsDsp = (nTsLi * 4) + nLi ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
      if(g_sqDspHdlcReceiver[i].nLi > 1) {
        DemoLogf("T4E:: Li%d is receive only\n", g_sqDspHdlcReceiver[i].nLi);
      }
      if(g_sqDspHdlcSender[i].nLi > 1) {
        DemoLogf("T4E:: Li%d is receive only\n", g_sqDspHdlcReceiver[i].nLi);
      }
    }

    // Connect pipes to the E1/T1 device
    memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
    if (g_demoEnv.b56kbpsChannel) {
      sqLiTimeSlotMask[nTsLi]   = 0xFE; // Timeslot on the T1/E1 span
      sqDspTimeSlotMask[nTsDsp] = 0xFE; // Timeslot on the DSP highway
    } else {
      sqLiTimeSlotMask[nTsLi]   = 0xFF; // Timeslot on the T1/E1 span
      sqDspTimeSlotMask[nTsDsp] = 0xFF; // Timeslot on the DSP highway
    }
    
    result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(result);
    
    result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
    
    // Connect pipes to the DSP device (Fixed timeslots since the Thor-2-ExpressCard does not have a time-space switch)
    result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
    
    // Create the following Connections in receive direction:
    //  Li  ------> HDLC Receiver
    result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(result);
    
    DemoLogf("Hdlc Receiver[%d] connected (Dsp#%d, Li#%d:TS%d) \n", i, g_sqDspHdlcReceiver[i].nDsp, g_sqDspHdlcReceiver[i].nLi, g_sqDspHdlcReceiver[i].nTs);
    
    
    result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(result);
    
    result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
    
    // Connect pipes to the DSP device (Fixed timeslots since Thor-2-ExpressCard does not have a time-space switch)
    result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
    
    // Make the following connections in transmit direction:
    // HDLC Sender   -----> Li
    result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(result);
    
    DemoLogf("Hdlc Sender[%d]   connected (Dsp#%d, Li#%d:TS%d) \n", i, g_sqDspHdlcSender[i].nDsp, g_sqDspHdlcSender[i].nLi, g_sqDspHdlcSender[i].nTs);
  }

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    
    // Initialize the Hdlc Sender 
    result = OtxHdlcSenderInit(
      g_sqDspHdlcSender[i].hHandle, // Handle to logical HDLC Sender Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      //OTX_HDLC_FFT_ALL_ONES,        // Fill character(s) to be send between HDLC frames
      1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(result);          
        
    // Initialize the Hdlc Receiver 
    result = OtxHdlcReceiverInit(
      g_sqDspHdlcReceiver[i].hHandle, // Handle to logical HDLC Receiver Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      //OTX_HDLC_FFT_ALL_ONES,        // Fill character(s) to be send between HDLC frames
      1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                  // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(result);          
    
    if (g_demoEnv.bSs7FisuFiltering) {
      DemoLogf("Setting FISU and LSSU filtering for device %d\n", i);
      result = OtxHdlcSS7SetFilter(g_sqDspHdlcReceiver[i].hHandle, 
        OTX_HDLC_SS7_FILTER_FISU | 
        OTX_HDLC_SS7_FILTER_SHORT_LSSU | 
        OTX_HDLC_SS7_FILTER_LONG_LSSU); 
      DemoCheckRc(result);
    }
    
  }
}



/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    result = OtxDrvEnable(g_sqDspHdlcSender[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
  
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000;
}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;

  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    //DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp2(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;

  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    fprintf(g_pFrameRxFile, "%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    //DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];

#if 0
  if (g_bHdlcVerbose) {
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
  }

  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }
#endif

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
       DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
      OTX_UINT8  buf[DEMO_HDLC_RX_BUFFER_SIZE];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nFrameSeq;
      OtxHdlcFrameEndCodeE nFrameEndCode;
      OTX_DATETIME nTimeStamp;
      OTX_RESULT result;
      OtxHdlcSS7FisuS *pFisu;
      OTX_INT32 nDev; 
      
      pFisu = 0; // Initialize to remove build warnings 

      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {

        // Log the received frames to a separate file
        DemoPrintTimeStamp2(pEventData->m_nTimeStamp);
        fprintf(g_pFrameRxFile, " %d, %d, %d, %d, ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
        for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
          fprintf(g_pFrameRxFile, "%.2X ", buf[i]);
        }
        fprintf(g_pFrameRxFile, "\n");

        if (pEventData->m_nSourceId < DEMO_NUMBER_OF_DEVS) {
          g_coHdlcRxFrames[pEventData->m_nSourceId]++;
          g_coHdlcRxBytes[pEventData->m_nSourceId] += nFrameLen;
          if (nFrameEndCode) g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
        } else {
          DemoLogf("HDLC Frame Received with an invalid m_nSourceId (%d)\n", pEventData->m_nSourceId);
        }

        nDev = pEventData->m_nSourceId;

        if (g_bHdlcVerbose) {
          char strBuf[1024];
          //DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
          
          sprintf(strBuf, "HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
          //DemoLogf("[%d] Len=%.3d, Fe=0x%.2X   ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode);
          for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
            sprintf(&strBuf[strlen(strBuf)], "%.2X ", buf[i]);
          }
          sprintf(&strBuf[strlen(strBuf)], "\n");

          DemoLogf(strBuf);
          
        }

        if (g_bSs7SimulationActive[nDev]) {
          if (nFrameEndCode == 0) {
            OTX_BOOL bErroredFrame = OTX_FALSE;
            
            pFisu = (OtxHdlcSS7FisuS *)&buf[0];
            
            if (((OTX_UINT8)pFisu->fsn != g_sqDspHdlcReceiver[nDev].nPrevFsn) && ((OTX_UINT8)pFisu->fsn != ((g_sqDspHdlcReceiver[nDev].nPrevFsn+1) % 128))) {
              bErroredFrame = OTX_TRUE;
              g_sqDspHdlcReceiver[nDev].nSs7Errors++;
              DemoPrintTimeStamp(pEventData->m_nTimeStamp);
              DemoLogf("=============v Fsn Error (%d) FSN=%d, prevFSN=%d\n", g_sqDspHdlcReceiver[nDev].nSs7Errors, pFisu->fsn, g_sqDspHdlcReceiver[nDev].nPrevFsn);
              DemoLogf("%d:Len=%.3d  ", pEventData->m_nSourceId, nFrameLen);
              for (i = 0; i < nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");

            }
            
            if (nFrameLen < 3) {
              DemoPrintTimeStamp(pEventData->m_nTimeStamp);
              DemoLogf("HDLC[%d] Received a too short frame\n", nDev);
            } else if (nFrameLen == 3) {
              // FISU, No action needed
            } else {
              // MSU, Acknowledge the frame by changing the FISU
              
              g_nSs7SimulationFramesReceived[nDev]++;


              OsaMutexLock(&g_demoSendSs7DataCritSec);

              if (bErroredFrame) {
                // Leave the BSN number the same, and toggle bib
                if (g_sqDspHdlcSender[nDev].fisu.bib) {
                  g_sqDspHdlcSender[nDev].fisu.bib = 0;
                } else {
                  g_sqDspHdlcSender[nDev].fisu.bib = 1;
                }

              } else {
                g_sqDspHdlcSender[nDev].fisu.bsn = pFisu->fsn;
                g_sqDspHdlcSender[nDev].fisu.bib = pFisu->fib;
              }
              g_sqDspHdlcSender[nDev].fisu.li = 0;
                            
              result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, (OTX_UINT8 *)&g_sqDspHdlcSender[nDev].fisu, 3, &g_sqDspHdlcSender[nDev].fisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 

              OsaMutexUnlock(&g_demoSendSs7DataCritSec);

              if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
                DemoLogf("HDLC[%d] Driver IO Request buffer is full. We need to retransmit this frame [not yet implemented]\n", nDev);
                // Driver IO Request buffer is full. We need to retransmit this frame.
              } else {
                if (OTX_FAILURE(result)) DemoLogf("HDLC[%d] OtxHdlcSS7SendData() failed. hHandle=%X\n", nDev, pEventData->m_hDevice);
                DemoCheckRc(result);
              }

              g_sqDspHdlcReceiver[nDev].nPrevFsn = pFisu->fsn;
  
            }
            
          } else {
            g_sqDspHdlcReceiver[nDev].nSs7Errors++;
            //DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            DemoLogf("%d: Frame with Bad FrameEndCode received (Len=%.3d  Fe=0x%.2X)\n", pEventData->m_nSourceId, nFrameLen, nFrameEndCode);
          }
          
        } else {
          g_sqDspHdlcReceiver[nDev].nSs7Errors++;
          //DemoPrintTimeStamp(pEventData->m_nTimeStamp);
          DemoLogf("%d:Len=%.3d  Unexpected frame. Simulation is not active on this device \n", pEventData->m_nSourceId, nFrameLen);
        }
      
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
      }
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(otxRc);
      }
    } else {
      // Unexpected event from this device
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;

  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}




/*******************************************************************************

  @func

  Transmits a number of MSU frames

  @comm

*******************************************************************************/
void *DemoSs7SimulationFunction(
  void *lpParameter   // thread data
)
{
  OTX_DATETIME nTimeStamp;
  OTX_RESULT result;
  OTX_INT32 i;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OTX_UINT8 sqHdlcFrame[300];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;
  OTX_INT32 coHdlcSenderFrames = g_demoEnv.nSimFrames;

  nDev = *((OTX_INT32 *) lpParameter);

  g_sqDspHdlcSender[nDev].busyCount = 0;

  g_sqDspHdlcSender[nDev].coMsusSent = 0;
  g_sqDspHdlcSender[nDev].coFisusSent = 0;

  g_sqDspHdlcSender[nDev].fisu.bib = 1;
  g_sqDspHdlcSender[nDev].fisu.fib = 1;
  g_sqDspHdlcSender[nDev].fisu.fsn = 0x7F;
  g_sqDspHdlcSender[nDev].fisu.bsn = 0x7F;
  g_sqDspHdlcSender[nDev].fisu.li = 0;
  g_sqDspHdlcSender[nDev].fisu.reserved = 0;
  

  g_bSs7SimulationActive[nDev] = OTX_TRUE;

  g_nSs7SimulationFramesReceived[nDev] = 0;

  srand(g_sqDspHdlcSender[nDev].hHandle);

  memcpy(sqHdlcFrame, &g_sqDspHdlcSender[nDev].fisu, OTX_HDLC_SS7_CB_FISU);  
  cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
  
  DemoLogf("%d: Sending %d MSU frames (Dsp#%d, Li#%d:TS%d)\n", nDev, coHdlcSenderFrames, g_sqDspHdlcSender[nDev].nDsp, g_sqDspHdlcSender[nDev].nLi, g_sqDspHdlcSender[nDev].nTs);

  result = OtxHdlcStartRepeatData(g_sqDspHdlcSender[nDev].hHandle, (OTX_UINT8 *) &g_sqDspHdlcSender[nDev].fisu, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(result);

  OsaThreadSleep(50);
  
  // Send MSU Frame(s)
  startTime = clock();
  while (g_sqDspHdlcSender[nDev].coMsusSent < coHdlcSenderFrames) {

    if (g_sqDspHdlcSender[nDev].coMsusSent == 0) {
      DemoLogf("%d: hHandle=%X\n", nDev, g_sqDspHdlcSender[nDev].hHandle);
    }

    OsaMutexLock(&g_demoSendSs7DataCritSec);

    // Step FSN for every new frame package (MSU+FISU or FISU+FISU) to send
    g_sqDspHdlcSender[nDev].fisu.fsn++;
    
    // Prepare an MSU
    //cbHdlcFrame = ((g_sqDspHdlcSender[nDev].coMsusSent%10)*10) + 10;
    if ((rand() % 10) < 3) {
      // Longer frames
      cbHdlcFrame = rand() % (OSA_MIN(272, g_demoEnv.cbMaxHdlcFrameLength)-4+1) + 4;
    } else {
      // Shorter frames
      cbHdlcFrame = rand() % (OSA_MIN(25, g_demoEnv.cbMaxHdlcFrameLength)-4) + 4;
    }
    memcpy(&msu, &g_sqDspHdlcSender[nDev].fisu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
    msu.li = 3;  // Set the Length Indicator; Anything larger than 2 is an MSU
    memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
    for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {     // Add the payload the HDLC frame buffer
      //sqHdlcFrame[i] = (OTX_UINT8) i;
      sqHdlcFrame[i] = (OTX_UINT8) rand();
    }

    result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, &g_sqDspHdlcSender[nDev].fisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 

    OsaMutexUnlock(&g_demoSendSs7DataCritSec);

    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      OtxDrvGetCurrentTime(&nTimeStamp);
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("%d: Driver IO Request buffer is full. We need to retransmit this frame. [not implemented]\n", nDev);
      g_sqDspHdlcSender[nDev].busyCount++;
      OsaThreadSleep(1);
    } else {
      DemoCheckRc(result);

      OsaThreadSleep(cbHdlcFrame/8 + rand() % 40);

      if (g_bHdlcVerbose) {
        char strBuf[1024];
        
        sprintf(strBuf, "%d: Sent MSU# %d (Len=%d  ", nDev, g_sqDspHdlcSender[nDev].coMsusSent, cbHdlcFrame);
        for (i = 0; i < cbHdlcFrame; i++) {
          sprintf(&strBuf[strlen(strBuf)], "%.2X ", sqHdlcFrame[i]);
        }
        sprintf(&strBuf[strlen(strBuf)], ")\n");
        
        DemoLogf(strBuf);
        
      } else {
        if ((g_sqDspHdlcSender[nDev].coMsusSent % 100) == 0) {
          DemoLogf("%d: Sent %d MSU frames so far\n", nDev, g_sqDspHdlcSender[nDev].coMsusSent);
        }
      }


      g_sqDspHdlcSender[nDev].coMsusSent++;
      
    }
  }
  stopTime = clock();
  
  DemoLogf("[%d] Done Sending (BusyCount=%d, time=%2.1f seconds) FISUs=%d MSUs=%d\n", nDev, g_sqDspHdlcSender[nDev].busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_sqDspHdlcSender[nDev].coFisusSent, g_sqDspHdlcSender[nDev].coMsusSent);
  DemoLogf("[%d] Press 's' to see stats of receiver and sender devices\n", nDev);
  //DemoLogf("We are done. Exiting thread.\n");

#if 1
  while (1) {
    OsaThreadSleep(500);
  }
#endif

  return 0;
}






/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bSs7FisuFiltering = OTX_TRUE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_2_EXPRESSCARD;
  envP->coLi              = 2;
  envP->coDsp             = 1;
  envP->bHdlcBitInversion = OTX_FALSE;
  envP->b56kbpsChannel    = OTX_FALSE;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLC_INTMEM;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  envP->nSimFrames        = DEMO_DEFAULT_SS7_SIM_FRAMES;
  envP->cbMaxHdlcFrameLength = DEMO_DEFAULT_HDLC_MAX_FRAME_LENGTH;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'S') {   // Sim frames
          envP->nSimFrames = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'H') {   // Max HDLC Frame Length
          envP->cbMaxHdlcFrameLength = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'I') {   // HDLC Bit Inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i]+1) == '5') {   // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   // Line Interface 1
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == 'E') {   // External
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   // Split clock, both Lis synchronized to its own line
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   // Split clock, Li0 synchronized to its T1E1 span. Li1 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   // Split clock, Li1 synchronized to its T1E1 span. Li0 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == 'G') {   // OtxHdlcEx_Intmem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC_INTMEM;
          } else if (*(argv[i]+2) == 'N') {   // OtxSpm2_Intmem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2_INTMEM;
          } else if (*(argv[i]+2) == 'M') {   // OtxSpm2Ex_Intmem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2EX_INTMEM;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coLi  = 2;
            envP->coDsp = 1;
          }else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coLi  = 4;
            envP->coDsp = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else if ( *(argv[i]+2) == 'J') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF_J1; // J1_ESF
          } else if ( *(argv[i]+2) == '5') {
            envP->eT1FrameFormat = OTX_T1_FF_F4_J1; 
          } else if ( *(argv[i]+2) == 'F') {
            envP->eT1FrameFormat = OTX_T1_FF_F12_J1; // J1_SF
          } else if ( *(argv[i]+2) == '0') {
            envP->eT1FrameFormat = OTX_T1_FF_F72_J1; // J1_SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>   : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[C]  : Board Type used for the demo, where\n");
  DemoLogf("                   C = Use Thor-2-ExpressCard\n");
  DemoLogf("                   R = Use Thor-4-ExpressCard\n");
  DemoLogf(" /L<fileName>  : Log file name\n");
  DemoLogf(" /S<number>    : Number of MSUs to send during the simulation\n");
  DemoLogf(" /H<number>    : Max HDLC (MSU) Frame length to use during the simulation\n");
  DemoLogf(" /ME           : E1 Li mode\n");
  DemoLogf(" /MT           : T1/J1 Li mode\n");
  DemoLogf(" /I            : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  DemoLogf(" /5            : Use HDLC 56kbps channel.\n");
  DemoLogf(" /P[G|N|M]     : DSP program to load, where\n");
  DemoLogf("                   G = OtxHdlc_IntMem \n");
  DemoLogf("                   N = OtxSpm2_IntMem \n");
  DemoLogf("                   M = OtxSpm2Ex_IntMem \n");
  DemoLogf(" /F[E|4|S|9|J|5|F|0] : T1/J1 Frame Format, where\n");
  DemoLogf("                   E = ESF (default)\n");
  DemoLogf("                   4 = F4\n");
  DemoLogf("                   S = SF [D3/D4] (F12)\n");
  DemoLogf("                   9 = SLC96 (F72)\n");
  DemoLogf("                   J = J1_ESF\n");
  DemoLogf("                   5 = J1_F4\n");
  DemoLogf("                   F = J1_SF\n");
  DemoLogf("                   0 = J1_SLC96\n");
  DemoLogf(" /J[A|B]       : T1/J1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]     : E1 Frame Format, where\n");
  DemoLogf("                   D = Doubleframe\n");
  DemoLogf("                   C = CRC4 Multiframe (default)\n");
  DemoLogf("                   G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]       : E1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   H = HDB3 (default)\n");
  DemoLogf(" /Z<str>       : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q            : Exit without prompting on errors\n");
  DemoLogf(" /K[I|0|1]     : Clock source, where\n");
  DemoLogf("                   I = Internal on-board oscillator\n");
  DemoLogf("                   0 = Recovered clock from Line Interface 0\n");
  DemoLogf("                   1 = Recovered clock from Line Interface 1\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    DemoLogf("\nExitOnError....\n");
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentHdlcDev = 0;
  OTX_RESULT result;

  DemoLogf("\nKey %c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" t   Start SS7 simulation on one channel on Li0 and Li1\n");
    DemoLogf(" T   Start SS7 simulation on all configured channels on Li0 and Li1\n");   
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" s   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
    DemoLogf(" p   Reset current HDLC Receiver\n");
    DemoLogf(" v   Toggle verbose logging of received HDLC frames\n");

    // Command applicable to the current HDLC device only
    DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" F   Toggle OTX_HDLC_SS7_FILTER_FISU_ALL filter mask on/off for current HDLC Sender\n");
    DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
    break;

  case 'x':
    {
      char szBuf[12];
      OTX_INT32 nDsp;                                                             
      
      DemoLogf("Dump profiling data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 3);
    }
    break;
  case 'j':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Dump debug data. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
    }
    break;
  case 'J':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Dump debug data for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;
  case 'k':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 0;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;
  case 'K':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = 1;
      DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;

  case 'y':
    {
      OTX_UINT32 nAttribValue;
      DemoLogf("Fifo Diagnostics for HDLC Receiver device#%d:\n", nCurrentHdlcDev);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_FRAMES, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Size (in frames)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_OCTETS, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Size (in octets)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_FRAMES, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Peak usage (in frames)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_OCTETS, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Peak usage (in octets)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_FRAMES, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Current usage (in frames): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_OCTETS, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Current usage (in octets): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_LOST_FRAMES, &nAttribValue); DemoCheckRc(result);
      DemoLogf("  Number of lost frames    : %d\n", nAttribValue);
    }
    break;


  case 'v':
    {
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
    }
    break;

  case 's':
    {
      OTX_INT32 i;
      OTX_UINT32  nFisuCount[DEMO_NUMBER_OF_DEVS];   // Number of FISUs received
      OTX_UINT32  nLssuCount[DEMO_NUMBER_OF_DEVS];   // Number of LSSUs received
      OTX_UINT32  nMsuCount[DEMO_NUMBER_OF_DEVS];    // Number of MSUs received
      OtxHdlcSS7FisuS fisu[DEMO_NUMBER_OF_DEVS];     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for HdlcReceiver #%d - #%d\n", 0, DEMO_NUMBER_OF_DEVS-1);

      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[i].hHandle, &nFisuCount[i], &nLssuCount[i], &nMsuCount[i], &fisu[i]); DemoCheckRc(result);
      }

      DemoLogf("FISUs received: ");
      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        DemoLogf("%.6lu ", nFisuCount[i]);
      }
      DemoLogf("\n");

      DemoLogf("LSSUs received: ");
      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        DemoLogf("%.6lu ", nLssuCount[i]);
      }
      DemoLogf("\n");

      DemoLogf("MSUs received:  ");
      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        DemoLogf("%.6lu ", nMsuCount[i]);
      }
      DemoLogf("\n");

      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        DemoLogf("Last FISU%d received: bsn=%d, bib=%d, fsn=%d, fib=%d, li=%d\n", i,
               fisu[i].bsn, fisu[i].bib, fisu[i].fsn, fisu[i].fib, fisu[i].li);
      }

      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
          DemoLogf("RX Tot[%.2d]=%.6d (Bad[%.2d]=%.6d)   TX MSUs=%d BusyCount=%d ERR=%d\n", 
            i, g_coHdlcRxFrames[i], 
            i, g_coHdlcBadRxFrames[i],
            g_sqDspHdlcSender[i].coMsusSent, 
            g_sqDspHdlcSender[i].busyCount,
            g_sqDspHdlcReceiver[i].nSs7Errors
            );
      }
      DemoLogf("\n");
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

  case 'd':
    DemoLogf("Dump DSP debug data\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 0);
    break;

  case 'b':
    DemoLogf("Init Bsp\n");
    OtxDrvDbgDump(g_hPhysDevs.hDsp[0], 1);
    break;


  case 't':
    {

      // This performs a simulation of MSU and FISU sending on the first two devices
      OTX_INT32 i;
      
      for (i = 0; i < 2; i++) {
        OTX_INT32 nRxDev = g_sqDspHdlcSender[i].nRxDev;
        
        g_sqDspHdlcReceiver[nRxDev].nPrevFsn   = 0x7F;
        g_sqDspHdlcReceiver[nRxDev].nSs7Errors = 0;
        g_coHdlcRxFrames[nRxDev]               = 0;
        g_coHdlcRxBytes[nRxDev]                = 0;
        g_coHdlcBadRxFrames[nRxDev]            = 0;
        g_bSs7SimulationActive[nRxDev]         = OTX_TRUE;
        
        DemoLogf("Starting SS7 Simulation. Sender dev #%d, Receiver dev #%d\n", i, nRxDev);
        
        OsaThreadStartEx(DemoSs7SimulationFunction, (void *) &g_sqDspHdlcSender[i].nDev);
      }
      
    }
    break;

  case 'T':
    {

      // This performs a simulation of MSU and FISU sending on all configured devices
      OTX_INT32 i;
      
      for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
        OTX_INT32 nRxDev = g_sqDspHdlcSender[i].nRxDev;
        
        g_sqDspHdlcReceiver[nRxDev].nPrevFsn   = 0x7F;
        g_sqDspHdlcReceiver[nRxDev].nSs7Errors = 0;
        g_coHdlcRxFrames[nRxDev]               = 0;
        g_coHdlcRxBytes[nRxDev]                = 0;
        g_coHdlcBadRxFrames[nRxDev]            = 0;
        g_bSs7SimulationActive[nRxDev]         = OTX_TRUE;
        
        DemoLogf("Starting SS7 Simulation. Sender dev #%d, Receiver dev #%d\n", i, nRxDev);
        
        OsaThreadStartEx(DemoSs7SimulationFunction, (void *) &g_sqDspHdlcSender[i].nDev);
      }
      
      
    }
    break;



  case 'F':
    {
      OTX_HANDLE hHandle;
      OTX_UINT32 filterMask;

      hHandle = g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle;

      result = OtxHdlcSS7GetFilter(hHandle, &filterMask); DemoCheckRc(result);
      if (filterMask & OTX_HDLC_SS7_FILTER_FISU_ALL) {
        filterMask &= ~OTX_HDLC_SS7_FILTER_FISU_ALL;
      } else {
        filterMask |= OTX_HDLC_SS7_FILTER_FISU_ALL;
      }

      DemoLogf("Setting SS7 filter mask to %.4X for HDLC Receiver #%d (Dsp#%d, Li#%d)...", filterMask, nCurrentHdlcDev, g_sqDspHdlcReceiver[nCurrentHdlcDev].nDsp, g_sqDspHdlcReceiver[nCurrentHdlcDev].nLi);

      result = OtxHdlcSS7SetFilter(hHandle, filterMask); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      printf("Set Clock Source:\n");
      printf("0 = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");
      printf("1 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to FA0 receive span\n");
      printf("2 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to FA1 receive span\n");

      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      case 1:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
        break;
      case 2:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
        break;
      default:
        printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        break;
      }

      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_H100, eClkSource); DemoCheckRc(result);

      // Setting the clock source also performs a reset of the E1T1 devices. Therefore we need to reconfigure the E1T1 devices here.
      DemoConfigureE1T1Devices();

      printf("Done.\n");
    }
    break;
  
  case 'n':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev++;
      if (nCurrentHdlcDev > (DEMO_NUMBER_OF_DEVS)-1) {
        // Wrap 
        nCurrentHdlcDev = 0;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev--;
      if (nCurrentHdlcDev < 0) {
        // Wrap
        nCurrentHdlcDev = (DEMO_NUMBER_OF_DEVS)-1;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
    }
    break;

  case 'I':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'O':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;


    
        
  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      printf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        printf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);
  OsaMutexInit(&g_demoSendSs7DataCritSec);

  DemoLogf("TecSs7, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("SS7 Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  if ((g_pFrameRxFile = fopen("TecSs7RxFrames.log", "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", "TecSs7RxFrames.log");
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  DemoInitializeGlobalData();

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  {
    char *pszClockSource;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      pszClockSource = "Internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
      pszClockSource = "Line Interface 0";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_1:
      pszClockSource = "Line Interface 1";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_2:
      pszClockSource = "External";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_3:
      pszClockSource = "Split Clock, both Line Interface are synchronized to its T1E1 span";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_4:
      pszClockSource = "Split Clock, Li0 synchronized to its T1E1 span. Li1 internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_5:
      pszClockSource = "Split Clock, Li1 synchronized to its T1E1 span. Li0 internal";
      break;
    default:
      pszClockSource = "Unknown";
      break;
    }

    DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
    DemoLogf("Clock Source: %s\n", pszClockSource);
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Set clock source 
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(result);

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszCpldBuildNumbers;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }
  }
  
  DemoLogf("\nPress 'h' for help\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

#if 1
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  if (exitFlag) {
    DemoLogf("\nexitFlag == TRUE.... Exiting\n");
  }
#else // Just for testing. Exiting automatically after 300ms
  OsaThreadSleep(300);
#endif

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
